# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    GUI Colors.

    Convenient named colors
"""

WHITE = 255, 255, 255
LIGHT_GRAY = 192, 192, 192
MEDIUM_GRAY = 128, 128, 128
DARK_GRAY = 64, 64, 64
BLACK = 0, 0, 0

RED = 255, 0, 0
YELLOW = 255, 255, 0
GREEN = 0, 255, 0
CYAN = 0, 255, 255
BLUE = 0, 0, 255
MAGENTA = 255, 0, 255

DARK_RED = 128, 0, 0
DARK_YELLOW = 128, 128, 0
DARK_GREEN = 0, 128, 0
DARK_CYAN = 0, 128, 128
DARK_BLUE = 0, 0, 128
DARK_MAGENTA = 128, 0, 128

# These work well with the LCARS color scheme
LCARS_RED = 206, 102, 102
LCARS_ORANGE = 248, 152, 29
LCARS_PURPLE = 204, 154, 198
LCARS_GRAY_BLUE = 155, 189, 228
LCARS_BEIGE = 255, 204, 153
LCARS_BLUE = 152, 152, 204
LCARS_PEACH = 255, 153, 102
LCARS_PINK = 204, 102, 153
